pub struct DiffSerializer<TradeOff: AnyTradeOff> { /* private fields */ }Expand description
All state-mutating function signatures that the code generator produces require this struct to be passed in. It follows the opaque handle pattern, so everything Just Works™. Internally, it’s capturing and serializing the before and after values, allowing the value to be both rolled back and transmitted to clients.
Examples
let diff: &mut DiffSerializer = &mut ctx.diff;
let mut cur_time = ctx.state.get_time();
cur_time -= 1;
ctx.state.set_time(cur_time, diff);Trait Implementations
Sourceimpl<TradeOff: Default + AnyTradeOff> Default for DiffSerializer<TradeOff>
impl<TradeOff: Default + AnyTradeOff> Default for DiffSerializer<TradeOff>
Sourcefn default() -> DiffSerializer<TradeOff>
fn default() -> DiffSerializer<TradeOff>
Returns the “default value” for a type. Read more