pub struct Physics {
pub integration_parameters: IntegrationParameters,
pub islands: IslandManager,
pub broad_phase: BroadPhaseBvh,
pub narrow_phase: NarrowPhase,
pub rigid_bodies: RigidBodySet,
pub colliders: ColliderSet,
pub impulse_joints: ImpulseJointSet,
pub multibody_joints: MultibodyJointSet,
pub ccd_solver: CCDSolver,
/* private fields */
}Expand description
Wrapper around all types required to step the Rapier simulation. Due to time constraints, the entire physics scene must be rebuilt at the start of every tick. All rigid bodies are discarded at the end of the tick. Will revisit someday to make this less awful.
Fields
integration_parameters: IntegrationParametersResets every tick
islands: IslandManagerResets every tick
broad_phase: BroadPhaseBvhResets every tick
narrow_phase: NarrowPhaseResets every tick
rigid_bodies: RigidBodySetResets every tick
colliders: ColliderSetResets every tick
impulse_joints: ImpulseJointSetResets every tick
multibody_joints: MultibodyJointSetResets every tick
ccd_solver: CCDSolverResets every tick
Implementations